#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Connection '''

from __future__ import print_function, division, absolute_import, unicode_literals

import base64
import collections
import os
import pandas as pd
import re
import requests
import six
import sys
import textwrap
import threading
import warnings
import xml.etree.ElementTree as ET
from numpy import nan
from six.moves import urllib
from .base import RESTHelpers, ESPObject
from .algorithm import Algorithm
from .config import get_option, ESP_ROOT, CONCAT_OPTIONS
from .connectorinfo import ConnectorInfo
from .mas import MASModule
from .router import Router
from . import contquery
from . import project
from . import windows
from .evtgen import EventGenerator
from .exceptions import ESPError
from .logger import Logger
from .metadata import Metadata
from .utils import xml
from .utils.authinfo import query_authinfo
from .utils.rest import get_params
from .utils.data import get_project_data, gen_name, get_server_info
from .utils.events import get_events
from .utils.keyword import dekeywordify
from .utils.project import expand_path
from .websocket import WebSocketClient
from .windows import BaseWindow, get_window_class


class ProjectStats(object):
    '''
    Project statistics subscriber

    Parameters
    ----------
    session : requests.Session or ESP or Project
        The object that supplies the ESP session
    filter : string, optional
        Functional filter to match projects.  If a project is given as
        the first argument, the filter is set to only view that project.
    interval : int, optional
        The interval in seconds between updates
    min_cpu : int, optional
        The minimum CPU value you want included
    limit : int, optional
        The maximum number of rows to retain in the DataFrame

    '''

    def __init__(self, session, filter=None, interval=None, min_cpu=None,
                 limit=20):
        self._ws = None
        self.filter = filter
        self.interval = interval
        self.min_cpu = min_cpu
        self.limit = limit

        if isinstance(session, project.Project):
            self.filter = "in(name,'%s')" % session.name
            self.session = session.session
        elif isinstance(session, ESPObject):
            self.session = session.session
        else:
            self.session = session

        stats = pd.DataFrame(columns=['project', 'contquery',
                                      'window', 'interval', 'cpu'],
                             data=[['', '', '', 0, 0.0]])
        stats = stats.set_index(['project', 'contquery', 'window'])
        self.stats = stats.iloc[0:0]

    def __getitem__(self, key):
        return self.stats[key]

    def __getattr__(self, name):
        return getattr(self.stats, name)

    def __len__(self):
        return len(self.stats)

    @property
    def url(self):
        '''
        Return the URL of the project statitics

        Returns
        -------
        string

        '''
        url_params = get_params(**{'format': 'xml',
                                   'interval': self.interval,
                                   'filter': self.filter,
                                   'minCpu': self.min_cpu})
        url_params = '&'.join(['%s=%s' % (k, v) for k, v in sorted(url_params.items())])
        return (re.sub(r'^\w+:', 'ws:', self.session.base_url) +
                'projectStats?' + url_params)

    @property
    def is_active(self):
        '''
        Is the web socket active?

        Returns
        -------
        bool

        '''
        return self._ws is not None

    def start(self):
        ''' Initialize the web socket and start it in its own thread '''
        if self._ws is not None:
            return

        def on_message(sock, message):
            if re.match(r'^\s*\w+\s*:\s*\d+\s*$', message):
                return

            if get_option('debug.events'):
                sys.stderr.write('%s\n' % message)

            rows = []

            elem = xml.from_xml(message)
            for proj in elem.findall('./project'):
                proj_name = proj.attrib['name']
                for cq in proj.findall('./contquery'):
                    cq_name = cq.attrib['name']
                    for win in cq.findall('./window'):
                        win_name = win.attrib['name']
                        attrib = dict(win.attrib)
                        attrib.pop('name')
                        row = dict(project=proj_name,
                                   contquery=cq_name,
                                   window=win_name)
                        for key, value in six.iteritems(attrib):
                            if re.match(r'^\s*\d+\s*$', value):
                                value = int(value)
                            elif re.match(r'^\s*\d*\.\d*\s*$', value):
                                value = float(value)
                            row[key] = value
                        rows.append(row)

            if rows:
                data = pd.DataFrame(columns=rows[0].keys(), data=rows)
                data = data.set_index(['project', 'contquery', 'window'])
                data = data[['interval'] + list(sorted(x for x in data.columns
                                                       if x != 'interval'))]
                data = pd.concat([self.stats, data], **CONCAT_OPTIONS)
                data = data.sort_index().sort_values(['interval']).tail(self.limit)
                self.stats = data

        if get_option('debug.requests'):
            sys.stderr.write('WEBSOCKET %s\n' % self.url)

        self._ws = WebSocketClient(self.url, on_message=on_message)
        self._ws.connect()

        ws_thread = threading.Thread(name='%s-%s' % (id(self), self.url),
                                     target=self._ws.run_forever)
        ws_thread.daemon = True
        ws_thread.start()

    def stop(self):
        ''' Stop processing events and close the web socket '''
        if self._ws is not None:
            self._ws.close()
            self._ws = None

    close = stop


class EngineMetadata(Metadata):
    ''' Metadata for ESP Engine '''

    def __metadata__(self):
        out = {}
        for item in self._get('engineMetadata').findall('.//meta'):
            self._set_metadata_from_xml(item, out)
        return out

    def __setitem__(self, key, value):
        self._put('engineMetadata/%s' % key, data=str(value))

    def __delitem__(self, key):
        self._delete('engineMetadata/%s' % key)


class ESP(RESTHelpers):
    '''
    ESP Connection

    Parameters
    ----------
    hostname : string, optional
        Hostname of ESP server or ESP connection URL
    port : int, optional
        Port number of ESP server.  This is not needed if
        a URL is used in the first parameter.
    username : string, optional
        Username for authentication
    password : string, optional
        Password for authentication.  If a password is specified,
        without a username, this is assumed to be an OAUTH token.
    protocol : string, optional
        The protocol to use: http or https.  This is not
        needed if a URL is used in the first parameter.
    ca_bundle : string, optional
        Path to the certificate bundle if using SSL
    authinfo : string, optional
        Path to the authinfo file containing authentication
        information.

    Attributes
    ----------
    metadata : dict
        ESP engine metadata
    server_info : dict
        ESP server information
    connector_info : dict
        Information about ESP connectors
    api_docs : dict
        Swagger JSON documentation

    Examples
    --------
    Start ESP connection

    >>> esp = esppy.ESP('http://esp-host.com:8080')
    >>> esp.server_info
    { ... }

    Create a project from a definition file

    >>> proj = esp.load_project('/path/to/model.xml')
    >>> proj.start()

    '''

    Project = project.Project
    Query = ContinuousQuery = contquery.ContinuousQuery

    def __init__(self, hostname=None, port=None, username=None,
                 password=None, protocol=None, ca_bundle=None,
                 authinfo=None, auth_obj=None):
        # Use environment variables as needed
        if hostname is None and get_option('hostname'):
            hostname = get_option('hostname')
        if port is None and get_option('port'):
            port = int(get_option('port'))
        if username is None and os.environ.get('ESPUSER'):
            username = os.environ['ESPUSER']
        if password is None and os.environ.get('ESPPASSWORD'):
            password = os.environ['ESPPASSWORD']
        if protocol is None and get_option('protocol'):
            protocol = get_option('protocol')

        # Set default protocol
        if not protocol:
            if ca_bundle:
                protocol = 'https'
            else:
                protocol = 'http'

        # Construct base URL
        if re.match('^https?://', hostname):
            base_url = hostname
            if not base_url.endswith('/'):
                base_url = '%s/' % base_url
            if not re.search(r'%s/' % ESP_ROOT, base_url):
                base_url = urllib.parse.urljoin(base_url, '%s/' % ESP_ROOT)
        else:
            base_url = '%s://%s:%s/%s/' % (protocol, hostname, port, ESP_ROOT)

        session = requests.Session()
        session.base_url = base_url

        # Set certificate verification
        if ca_bundle:
            session.verify = ca_bundle

        # Use specified auth object
        if auth_obj is not None:
            session.auth = auth_obj

        else:
            enable_kerberos = username is None and password is None

            # Check for authinfo file
            if password is None:
                authinfo = query_authinfo(host=hostname, user=username,
                                          protocol=port, path=authinfo)
                if authinfo is not None:
                    username = authinfo.get('user', username)
                    password = authinfo.get('password')

            # Set authorization headers
            if username and password:
                session.headers.update({
                    'Authorization': b'Basic ' +
                                     base64.b64encode(('%s:%s' %
                                                       (username, password)
                                                       ).encode('utf-8')).strip()
                })

            elif password:
                session.headers.update({'Authorization': b'Bearer ' +
                                                         password.encode('utf-8')})

            # Verify authentication
            res = session.head(session.base_url)
            if res.status_code == 401 and enable_kerberos:
                try:
                    from requests_kerberos import HTTPKerberosAuth, OPTIONAL
                    session.auth = HTTPKerberosAuth(mutual_authentication=OPTIONAL)
                    res = session.head(session.base_url)
                    if res.status_code == 401:
                        raise ESPError(res.reason)

                except ImportError:
                    warnings.warn('Kerberos authentication attempted, but requests_kerberos '
                                  'module is not installed ', RuntimeWarning)
                    raise ESPError(res.reason)

        RESTHelpers.__init__(self, session=session)

        if float(self.server_info['version']) < 5.2:
            raise RuntimeError('This package requires an ESP server '
                               'version 5.2 or greater')

        self._populate_algorithms()

    def __str__(self):
        return '%s(%s)' % (type(self).__name__,
                           repr(self.base_url.replace('/%s/' % ESP_ROOT, '')))

    def __repr__(self):
        return str(self)

    def _enable_http_debugging(self):
        import logging
        import http.client
        http.client.HTTPConnection.debuglevel = 1
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger('requests.packages.urllib3')
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

    @property
    def metadata(self):
        ''' Engine metadata '''
        return EngineMetadata(session=self.session)

    def create_project(self, proj, **kwargs):
        '''
        Create a new Project object

        Notes
        -----
        This method does not add the project to the server.  The
        returned object is client-side only until you install the project
        using :meth:`load_project`.

        Parameters
        ----------
        proj : string
            Either the name of the project, or a filename / URL
            pointing to a project definition
        kwargs : keyword arguments
            Optional parameters to the Project constructor.
            These are only used in the `project` argument is a
            project name.

        See Also
        --------
        :meth:`load_project`

        Returns
        -------
        :class:`Project`

        '''
        if os.path.isfile(proj) or re.match(r'\w+://', proj):
            out = project.Project.from_xml(get_project_data(proj))
        elif proj.startswith('<') and proj.endswith('>'):
            out = project.Project.from_xml(proj)
        else:
            out = project.Project(proj, **kwargs)
        out.session = self.session
        return out

    def _populate_algorithms(self):
        ''' Generate algorithm classes '''
        for alg in ['train', 'calculate', 'score']:
            doc = []
            doc.append('%s Algorithms' % alg.title())
            doc.append('-' * len(doc[-1]))
            members = {}
            algorithms = []
            for key, value in sorted(self.get_algorithms(alg, properties=True).items()):
                algorithms.append(key)
                members[str(key)] = self._new_algorithm(value,
                                                        get_window_class('window-%s' %
                                                                         alg))
                doc.append(key)
            members['__doc__'] = '\n'.join(doc)
            members['supported_algorithms'] = tuple(algorithms)
            setattr(self, alg,
                    type(str('%s%s' % (alg.title(), 'Algorithms')),
                         (object,), members))

    def _new_algorithm(self, data, wcls):
        '''
        Create a new algorithm window class

        Parameters
        ----------
        data : Algorithm
            The algorithm class populated with information from the server
        wcls : Window class
            The window class to use as the super class

        Returns
        -------
        :class:`Window` subclass

        '''
        _globals = globals()
        _locals = locals()

        has_input_map = bool(getattr(data, 'input_map', None))
        has_output_map = bool(getattr(data, 'output_map', None))

        params = ', '.join(['%s=%s' % (dekeywordify(name),
                                       repr(value.get('default', None)))
                            for name, value in sorted(data.parameters.items())])
        setters = [('self.parameters["%s"] = '
                    'self._cast_parameter(%s, dtype=%s, vartype=%s)')
                   % (dekeywordify(name),
                      dekeywordify(name),
                      repr(value.get('type')), repr(value.get('vartype')))
                   for name, value in sorted(data.parameters.items())]

        doc = [re.sub(r'([A-Z]\w+)', lambda x: ' ' + x.group(1), data.name).strip()]

        doc.append('')

        doc.append('Parameters')
        doc.append('----------')
        doc.append('name : string, optional')
        doc.append('    The name of the window')

        for name, value in sorted(data.parameters.items()):
            name = dekeywordify(name)
            desc = value.get('description', '').capitalize().strip()

            dtype = value.get('type', '')
            if dtype.endswith('-list'):
                dtype = 'list-of-%ss' % dtype.replace('-list', '')

            doc.append('%s : %s, optional' % (name, value['type']))

            if desc:
                doc.append('    %s' % '\n    '.join(textwrap.wrap(desc, 76)))

        if has_input_map:
            doc.append('input_map : dict, optional')
            doc.append('    The input variable map entries')

        if has_output_map:
            doc.append('output_map : dict, optional')
            doc.append('    The output variable map entries')

        doc.append('')

        doc.append('See Also')
        doc.append('--------')
        doc.append('set_inputs')
        doc.append('set_outputs')

        doc.append('')

        input_map_types = {}
        input_map_doc = []

        if wcls.__name__.lower().startswith('calc'):
            if params:
                params += ', index_type=None, produces_only_inserts=None'
            else:
                params = 'index_type=None, produces_only_inserts=None'

        if has_input_map:

            doc.append('Notes')
            doc.append('-----')
            doc.append('** Input Map Entries **')
            doc.append('')

            input_map_doc.extend(['Set the input map entries', '',
                                  'Parameters', '----------'])

            for key, value in sorted(data.input_map.items()):
                key = dekeywordify(key)
                desc = value.get('description', '').capitalize().strip()
                dtype = value.get('type', '')
                dtype = dtype.replace('varlist', 'list-of-strings')
                dtype = dtype.replace('var', 'string')
                vartype = value.get('vartype', '')

                if vartype:
                    input_map_types[key] = vartype

                docstring = '%s : %s' % (key, dtype)
                if value.get('default', '').strip():
                    docstring = '%s, optional (default: %s)' % \
                                (docstring, value['default'])

                doc.append(docstring)
                input_map_doc.append(docstring)

                if desc:
                    doc.append(
                        '    %s' % '\n    '.join(textwrap.wrap(desc, 76)))
                    input_map_doc.append(
                        '    %s' % '\n        '.join(textwrap.wrap(desc, 72)))

            doc.append('')
            input_map_doc.append('')

            if params:
                params += ', input_map=None'
            else:
                params = 'input_map=None'

        output_map_types = {}
        output_map_doc = []

        if has_output_map:

            if 'Notes' not in doc:
                doc.append('Notes')
                doc.append('-----')

            doc.append('** Output Map Entries **')
            doc.append('')

            output_map_doc.extend(['Set the output map entries', '',
                                   'Parameters', '----------'])

            for key, value in sorted(data.output_map.items()):
                key = dekeywordify(key)
                desc = value.get('description', '').capitalize().strip()
                dtype = value.get('type', '')
                dtype = dtype.replace('varlist', 'list-of-strings')
                dtype = dtype.replace('var', 'string')
                vartype = value.get('vartype', '')

                if vartype:
                    output_map_types[key] = vartype

                docstring = '%s : %s' % (key, dtype)
                if value.get('default', '').strip():
                    docstring = '%s, optional (default: %s)' % \
                                (docstring, value['default'])

                doc.append(docstring)
                output_map_doc.append(docstring)

                if desc:
                    doc.append(
                        '    %s' % '\n    '.join(textwrap.wrap(desc, 76)))
                    output_map_doc.append(
                        '    %s' % '\n        '.join(textwrap.wrap(desc, 72)))

            doc.append('')
            output_map_doc.append('')

            if params:
                params += ', output_map=None'
            else:
                params = 'output_map=None'

        schema = ''
        if not wcls.__name__.lower().startswith('train'):
            schema = ', schema=schema'

        copyvars = ''
        if wcls.__name__.lower().startswith('score'):
            copyvars = ', copyvars=copyvars'

        six.exec_('\n'.join([
            '''def __init__(self, name=None%s%s, pubsub=None, %s):''' %
                (schema.replace('=schema', '=None'),
                 copyvars.replace('=copyvars', '=None'), params),
            '''    from .windows import %s''' % wcls.__name__,
            '''    %s.__init__(self, name=name%s%s, pubsub=pubsub)''' %
                (wcls.__name__, schema, copyvars),
            '''    self.algorithm = %s''' % repr(data.name),
            has_input_map and '''    self.set_inputs(**(input_map or {}))''' or '',
            has_output_map and '''    self.set_outputs(**(output_map or {}))''' or '',
            '''    %s''' % '\n    '.join(setters),
        ]), _globals, _locals)

        # Setup input_map methods
        input_map = ', '.join(['%s=%s' % (dekeywordify(x),
                                          dekeywordify(x))
                               for x in sorted(data.input_map.keys())])
        if input_map:
            input_map = ', ' + input_map

        alg = ''
        if wcls.__name__.lower().startswith('score'):
            alg = ', self.algorithm'

        six.exec_('\n'.join([
            '''def set_inputs(self%s):''' % re.sub(r'=\w+', r'=None', input_map),
            '''    """''',
            '    ' + '\n    '.join(input_map_doc),
            '''    """''',
            '''    from .windows import %s''' % wcls.__name__,
            '''    return %s.set_inputs(self%s%s)''' % (wcls.__name__, alg, input_map),
        ]), _globals, _locals)

        # Setup output_map methods
        output_map = ', '.join(['%s=%s' % (dekeywordify(x),
                                           dekeywordify(x))
                                for x in sorted(data.output_map.keys())])
        if output_map:
            output_map = ', ' + output_map

        six.exec_('\n'.join([
            '''def set_outputs(self%s):''' % re.sub(r'=\w+', r'=None', output_map),
            '''    """''',
            '    ' + '\n    '.join(output_map_doc),
            '''    """''',
            '''    from .windows import %s''' % wcls.__name__,
            '''    return %s.set_outputs(self%s%s)''' % (wcls.__name__, alg, output_map),
        ]), _globals, _locals)

        cls_members = {
            '__doc__': '\n'.join(doc),
            '__init__': _locals['__init__'],
            'output_map_types': output_map_types,
            'input_map_types': input_map_types,
            'set_inputs': _locals['set_inputs'],
            'set_outputs': _locals['set_outputs'],
        }

        if wcls.__name__.lower().startswith('train'):
            cls_members.pop('set_outputs')

        return type(str(data.name), (wcls,), cls_members)

    def _projects_from_xml(self, data):
        ''' Create projects from XML content '''
        projects = dict()
        for item in data.findall('./project'):
            proj = project.Project.from_xml(item, session=self.session)
            projects[proj.name] = proj
        return projects

    def get_projects(self, name=None, filter=None):
        '''
        Retrieve projects from the server

        Parameters
        ----------
        name : string or list-of-strings, optional
            Names of the projects to return
        filter : string or list-of-strings, optional
            Functional filter which specifies projects to return

        Returns
        -------
        dict of :class:`Project`

        '''
        out = self._get('projectXml', params=get_params(name=name, filter=filter))
        return self._projects_from_xml(out)

    def get_project(self, name):
        '''
        Retrieve the specified project

        Parameters
        ----------
        name : string
            Name of the project

        Returns
        -------
        :class:`Project`

        '''
        out = self.get_projects(name=name)
        try:
            out = out[name]
            out._load_metadata()
            return out
        except KeyError:
            raise KeyError("No project with name '%s' found." % name)

    def load_project(self, project, name=None, overwrite=True, start=True,
                     start_connectors=True):
        '''
        Load a project from a project definition

        Parameters
        ----------
        project : string or Project or file-like
            If a Project object, the project is exported to an XML definition
            then loaded.  If it is a string containing a URL, the URL must
            contain an XML definiton of a project.  Any other string or
            file-like object, must contain XML defining a project.
        name : string, optional
            Name of the project.  By default, the name is retrieved from
            the XML definition.
        overwrite : bool, optional
            Should an existing project with the same name be overwritten?
        start : bool, optional
            Should the project be started?
        start_connectors : bool, optional
            Should the connectors be started?

        Returns
        -------
        :class:`Project`

        '''
        if isinstance(project, six.string_types) and re.match(r'\w+://', project):
            data = ''
            project_url = project
        else:
            data = get_project_data(project)
            project_url = None

        if name is None:
            if data:
                proj = xml.from_xml(data)
                if proj.tag == 'project':
                    name = proj.attrib.get('name')
                else:
                    name = proj.findall('.//project')[0].attrib.get('name')
            if name is None:
                name = gen_name(prefix='p_')

        self._put('projects/%s' % name,
                  params=get_params(overwrite=overwrite,
                                    connectors=start_connectors,
                                    projectUrl=project_url,
                                    start=start),
                  data=data.encode('utf-8'))

        return self.get_project(name)

    install_project = load_project

    def delete_projects(self, *name, **kwargs):
        '''
        Delete projects

        Parameters
        ----------
        *name : zero-or-more-strings, optional
            Names of the projects to delete
        filter : string or list-of-strings, optional
            Functional filter indicating the projects to delete

        '''
        self._delete('projects', params=get_params(name=list(name),
                                                   filter=kwargs.get('filter')))

    def delete_project(self, name):
        '''
        Delete specified project

        Parameters
        ----------
        name : string
            Name of the project to delete

        '''
        self._delete('projects/%s' % name)

    def get_running_projects(self, name=None, filter=None):
        '''
        Retrieve running project information

        Parameters
        ----------
        name : string or list-of-strings, optional
            Names of the running projects to return
        filter : string or list-of-strings, optional
            Functional filters indicating the running projects to return

        Returns
        -------
        dict of :class:`Project`

        '''
        out = self._get('runningProjects', params=get_params(name=name,
                                                             filter=filter,
                                                             schema=False))
        names = list(self._projects_from_xml(out).keys())
        if names:
            return self.get_projects(name=names)

        return {}

    def get_running_project(self, name):
        '''
        Retrieve the specified running project

        Parameters
        ----------
        name : string
            Name of the project

        Returns
        -------
        :class:`Project`

        '''
        out = self.get_running_projects(name=name)
        if out:
            return out.popitem()[-1]
        raise KeyError("No running project with name '%s' found." % name)

    def start_projects(self, name=None, filter=None):
        '''
        Start projects

        Parameters
        ----------
        name : string or list-of-strings, optional
            Names of projects to start
        filter : string or list-of-strings, optional
            Functional filters indicating which projects to start

        '''
        self._post('runningProjects', params=get_params(name=name, filter=filter))

    def start_project(self, name):
        '''
        Start specified project

        Parameters
        ----------
        name : string
            Name of the project to start

        '''
        self._post('runningProjects/%s' % name, params=get_params(name=name))

    def get_stopped_projects(self, name=None, filter=None):
        '''
        Retrieve stopped projects

        Parameters
        ----------
        name : string or list-of-strings, optional
            Names of stopped projects to return
        filter : string or list-of-strings, optional
            Functional filters indicating stopped projects to return

        Returns
        -------
        dict of :class:`Project`

        '''
        out = self._get('stoppedProjects', params=get_params(name=name,
                                                             filter=filter,
                                                             schema=False))
        names = list(self._projects_from_xml(out).keys())
        if names:
            return self.get_projects(name=names)

        return {}

    def get_stopped_project(self, name):
        '''
        Retrieve specified stopped project

        Parameters
        ----------
        name : string
            Name of the stopped project

        Returns
        -------
        :class:`Project`

        '''
        out = self.get_stopped_projects(name=name)
        if out:
            return out.popitem()[-1]
        raise KeyError("No stopped project with name '%s' found." % name)

    def stop_projects(self, name=None, filter=None):
        '''
        Stop projects

        Parameters
        ----------
        name : string or list-of-strings, optional
            Names of the projects to stop
        filter : string or list-of-strings, optional
            Functional filters indicating which projects to stop

        '''
        self._post('stoppedProjects', params=get_params(name=name, filter=filter))

    def stop_project(self, name):
        '''
        Stop specified project

        Parameters
        ----------
        name : string
            Name of the project to stop

        '''
        self._post('stoppedProjects/%s' % name, params=get_params(name=name))

    def get_windows(self, path='*.*.*', type=None, filter=None):
        '''
        Retrieve windows from the server

        Parameters
        ----------
        path : string, optional
            '.' delimited path to windows to retrieve.  '*' can be used as
            a wildcard for any component.  Multiple component names can
            be delimited by '|'.
        type : string or list-of-strings, optional
            Types of windows you want to retrieve
        filter : string or list-of-strings, optional
            Function filter indicating which windows to retrieve

        Returns
        -------
        dict of :class:`Window`

        '''
        path = [None, None, None] + expand_path(path)
        res = self._get('windowXml', params=get_params(project=path[-3],
                                                       contquery=path[-2],
                                                       name=path[-1], type=type,
                                                       filter=filter))
        windows = dict()
        for item in res.findall('./*'):
            try:
                wcls = get_window_class(item.tag)
            except KeyError:
                raise TypeError('Unknown window type: %s' % item.tag)
            window = wcls.from_xml(item, session=self.session)
            windows[window.fullname] = window
        return windows

    def get_window(self, path):
        '''
        Retrieve specified window

        Parameters
        ----------
        path : string, optional
            '.' delimited path to windows to retrieve.  '*' can be used as
            a wildcard for any component.  Multiple component names can
            be delimited by '|'.

        Returns
        -------
        :class:`Window`

        '''
        out = self.get_windows(path)
        if len(out) > 1:
            raise ValueError(("More than one window with the path '%s' exists. " % path) +
                             'Use a more explicit path or use the get_windows method.')
        if out:
            return out.popitem()[1]
        raise KeyError("No window with the path '%s'" % path)

    def get_events(self, window_filter=None, event_filter=None, sort_by=None, limit=None):
        '''
        Retrieve events from the server

        Parameters
        ----------
        window_filter : string or list-of-strings, optional
            Functional filter indicating windows of which you want the
            events for
        event_filter : string or list-of-strings, optional
            Functional filter indicating events to retrieve
        sort_by : string, optional
            The field to sort results on.  It should be of the
            form: ``sortfield:ascending``.  Where the ordering can be
            ``ascending`` or ``descending``.
        limit : int, optional
            Maximum number of events to return

        Returns
        -------
        dict of :class:`pandas.DataFrame`

        '''
        return get_events(self, self._get('events',
                                          params=get_params(window_filter=window_filter,
                                                            event_filter=event_filter,
                                                            sort_by=sort_by,
                                                            limit=limit)))

    def get_pattern_events(self, window_filter=None, event_filter=None,
                           sort_by=None, limit=None):
        '''
        Retrieve events residing in open patterns in the server

        Parameters
        ----------
        window_filter : string or list-of-strings, optional
            Functional filter indicating windows of which you want the
            events for
        event_filter : string or list-of-strings, optional
            Functional filter indicating events to retrieve
        sort_by : string, optional
            The field to sort results on.  It should be of the
            form: ``sortfield:ascending``.  Where the ordering can be
            ``ascending`` or ``descending``.
        limit : int, optional
            Maximum number of events to return

        Returns
        -------
        dict of :class:`pandas.DataFrame`

        '''
        return get_events(self, self._get('patternEvents',
                                          params=get_params(window_filter=window_filter,
                                                            event_filter=event_filter,
                                                            sort_by=sort_by,
                                                            limit=limit)))

    @property
    def server_info(self):
        ''' Retrive current status of the server '''
        return get_server_info(self)

    def save(self, path=None):
        '''
        Save the server state

        Parameters
        ----------
        path : string, optional
            Location of the server data

        '''
        self._put('server/state', params=get_params(value='persisted', path=path))

    def reload(self):
        ''' Reset the server model to the model used at startup '''
        self._put('server/state', params=get_params(value='reloaded'))

    def shutdown(self):
        ''' Shutdown the server '''
        self._put('server/state', params=get_params(value='stopped'))

    def get_loggers(self):
        '''
        Retrieve all loggers in the system

        Returns
        -------
        dict of :class:`Logger`

        '''
        loggers = dict()
        out = self._get('loggers')
        for item in out.findall('./logger'):
            logger = Logger(**item.attrib)
            logger.session = self.session
            loggers[logger.name] = logger
        return loggers

    def get_logger(self, name):
        '''
        Retrieve the specified logger

        Parameters
        ----------
        name : string
            Name of the logger

        Returns
        -------
        :class:`Logger`

        '''
        try:
            logger = None

            out = self._get('loggers/%s' % name)
            node = out.find("logger")
            if node != None:
                logger = Logger(**node.attrib)
                logger.session = self.session

            return logger

        except ESPError:
            raise KeyError("No logger with the name '%s' exists." % name)

    @property
    def connector_info(self):
        '''
        Return the connection metadata for all connectors

        Returns
        -------
        dict

        '''
        out = {}
        info = self._get('connectorInfo')
        if info.tag == 'connector':
            conn = ConnectorInfo.from_xml(info, session=self.session)
            out[conn.label] = conn
        else:
            for connector in info.findall('./connector'):
                conn = ConnectorInfo.from_xml(connector, session=self.session)
                out[conn.label] = conn
        return out

    def _run_project(self, project, windows=None):
        '''
        Run a project and send back window contents

        Parameters
        ----------
        project : Project or string
            The project definition, project XML, or URL of project definition
        windows : string or list-of-strings, optional
            Names of the windows whose contents to include

        Returns
        -------
        dict

        '''
        if isinstance(project, six.string_types) and re.match(r'\w+://', project):
            data = ''
            project_url = project
        else:
            data = get_project_data(project)
            project_url = None

        return get_events(self, self._post('projectResults',
                                            params=get_params(windows=windows,
                                                              projectUrl=project_url),
                                            data=data))

    def validate_project(self, project):
        '''
        Validate project definition

        Parameters
        ----------
        project : string
            XML content or location of the project definition

        Returns
        -------
        True
            If project is valid
        False
            If project is not valid

        '''
        try:
            res = self._post('projectValidationResults', data=get_project_data(project))
            if res.tag != 'schema-validation-success':
                return False
        except ESPError:
            return False
        return True

    def get_mas_modules(self, expandcode=False):
        '''
        Retrieve all MAS modules

        Parameters
        ----------
        expandcode : bool, optional
            Should code-file elements be returned as code elements containing
            the contents of the file?

        Returns
        -------
        dict of :class:`MASModule`

        '''
        res = self._get('masModules', params=get_params(expandcode=expandcode))
        out = {}
        for elem in res.findall('./project'):
            proj = elem.attrib['name']
            modules = []
            for mod in elem.findall('./mas-module'):
                mod = MASModule.from_xml(mod, session=self.session)
                mod.project = proj
                modules.append(mod)
            out[proj] = modules
        return out

    def get_routers(self):
        '''
        Retrieve all routers in the server

        Returns
        -------
        dict of :class:`Router`

        '''
        res = self._get('routers')
        out = {}
        for elem in res.findall('./esp-router'):
            rte = Router.from_xml(elem, session=self.session)
            out[rte.name] = rte
        return out

    def get_router(self, name):
        '''
        Retrieve the specified router

        Parameters
        ----------
        name : string
            Name of the router

        Returns
        -------
        :class:`Router`

        '''
        return Router.from_xml(self._get('routers/%s' % name),
                               session=self.session)

    def get_router_stats(self, name=None):
        '''
        Retrieve router statistics

        Parameters
        ----------
        name : string, optional
            Name of the router to retrieve statistics for

        Returns
        -------
        dict

        '''
        if name:
            stats = self._get('routerStats/%s' % name)
            elems = [stats]
        else:
            stats = self._get('routerStats')
            elems = stats.findall('./esp-router')

        def cast_numeric(val):
            ''' Cast numeric looking values to numerics '''
            if re.match(r'^\d+$', val):
                return int(val)
            elif re.match(r'^[\d\.]+$', val):
                return float(val)
            return val

        out = {}

        for elem in elems:
            routes = {}
            for route in elem.findall('./route'):
                rte = {}
                for stat in route.findall('./stats/*'):
                    rte[stat.tag] = cast_numeric(stat.text)
                routes[route.attrib['name']] = rte
            out[elem.attrib['name']] = routes

        if name:
            return out[name]

        return out

    def create_router(self, name=None, overwrite=True):
        '''
        Create a new router

        Parameters
        ----------
        router : string
            The router definition
        name : string, optional
            The name of the router to create
        overwrite : bool, optional
            Should an existing router of the same name be overwritten?

        Returns
        -------
        :class:`Router`

        '''
        rte = Router(name=name)
        rte.session = self.session
        return rte

    def get_event_generators(self):
        '''
        Retrieve all event generators in the server

        Returns
        -------
        dict of :class:`EventGenerator`

        '''
        res = self._get('eventGenerators')
        out = {}
        for elem in res.findall('./event-generator'):
            evtg = EventGenerator.from_xml(elem, session=self.session)
            out[evtg.name] = evtg
        return out

    def get_event_generator_state(self, name):
        '''
        Retrieve the state of the specified event generator

        Parameters
        ----------
        name : string
            Name of the event generator

        Returns
        -------
        string

        '''
        return self._get('eventGenerators/%s' % name).attrib['state']

    def create_event_generator(self, window, data=None, name=None, overwrite=False):
        '''
        Create an event generator

        Parameters
        ----------
        window : Window
            The window to inject into
        data : CSV or URL or DataFrame, optional
            The events to inject
        name : string, optional
            The name of the event generator.  If no name is given,
            a name will be generated.
        overwrite : bool, optional
            Should an existing event generator with the same name
            be overwritten?

        Returns
        -------
        :class:`EventGenerator`

        '''
        gen = EventGenerator(name=name)
        gen.session = self.session
        gen.publish_target = window
        if data is not None:
            gen.event_data = data
            gen.save(overwrite=overwrite)
        return gen

    def delete_event_generators(self):
        ''' Delete all event generators in the server '''
        for key in self.get_event_generators():
            self.delete_event_generator(key)

    def delete_event_generator(self, name):
        '''
        Delete the specified event generator

        Parameters
        ----------
        name : string
            Name of the event generator to delete

        '''
        self._delete('eventGenerators/%s' % name)

    @property
    def api_docs(self):
        ''' Swagger JSON documentation '''
        return self._get('api-docs', format='json', headers={'accept':'application/json'})

    def get_algorithms(self, atype, properties=False, type=None, reference=None):
        '''
        Retrieve information for all algorithms of the specified type

        Parameters
        ----------
        atype : string
            Type of the algorithm: train, calculate, or score
        properties : bool, optional
            Should algorithm properties be returned?
        type : string, optional
            Set to 'astore' if astore information is desired
        reference : string, optional
            Set to the value of an astore reference file

        Returns
        -------
        dict of :class:`Algorithm`

        '''
        res = self._get('algorithms/%s' % atype,
                        params=get_params(properties=properties,
                                          type=type,
                                          reference=reference))
        out = {}
        if res.attrib.get('reference', None):
            alg = Algorithm.from_xml(res)
            out[alg.reference] = alg
        else:
            for item in res.findall('./algorithm'):
                alg = Algorithm.from_xml(item)
                out[alg.name] = alg
        return out

    def get_algorithm(self, atype, name):
        '''
        Retrieve information for the specified algorithm

        Parameters
        ----------
        atype : string
            Type of the algorithm: train, calculate, or score
        name : string
            Name of the algorithm

        Returns
        -------
        :class:`Algorithm`

        '''
        return Algorithm.from_xml(self._get('algorithms/%s/%s' %
                                            (atype, name)).find('./algorithm'))

    def get_project_stats(self, filter=None, interval=None, min_cpu=None, limit=20):
        '''
        Project statistics subscriber

        Parameters
        ----------
        filter : string, optional
            Functional filter to match projects.
        interval : int, optional
            The interval in seconds between updates
        min_cpu : int, optional
            The minimum CPU value you want included
        limit : int, optional
            The maximum number of rows to retain in the DataFrame

        Returns
        -------
        :class:`ProjectStats`

        '''
        proj_stats = ProjectStats(self, filter=filter, interval=interval,
                                  min_cpu=min_cpu, limit=limit)
        proj_stats.start()
        return proj_stats

    def enable_server_log_capture(self):
        ''' Enable capturing of server log messages '''
        self._put('logCapture/state', params=get_params(value='on'))

    def disable_server_log_capture(self):
        ''' Disable capturing of server log messages '''
        self._put('logCapture/state', params=get_params(value='off'))

    def get_server_log_state(self):
        '''
        Return the current state of the server log capture

        Returns
        -------
        string

        '''
        return self._get('logCapture').attrib['logcapture'] == 'on'

    def get_server_log(self):
        '''
        Return all stored server log messages

        Returns
        -------
        list of strings

        '''
        return self._get('logs', raw=True).rstrip().split('\n')


def get_subclasses(cls):
    for subclass in cls.__subclasses__():
        for subcls in get_subclasses(subclass):
            yield subcls
        yield subclass


for cls in get_subclasses(windows.BaseWindow):
    setattr(ESP, cls.__name__, cls)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
Utilities for testing

'''

from __future__ import print_function, division, absolute_import, unicode_literals

import copy
import difflib
import os
import re
import sys
import unittest
import six
from ..utils import xml


class TestCase(unittest.TestCase):
    ''' TestCase with SWAT Customizations '''

    def assertContentsEqual(self, file1, file2):
        if os.path.isfile(file1):
            with open(file1, 'rb') as infile:
                file1 = infile.read().decode('utf-8').split('\n')
        else:
            file1 = file1.split('\n')

        if os.path.isfile(file2):
            with open(file2, 'rb') as infile:
                file2 = infile.read().decode('utf-8').split('\n')
        else:
            file1 = file1.split('\n')

        out = list(difflib.Differ().compare(file1, file2) )
        print(out)
        if out:
            for line in out:
                print(line)
            raise ValueError('Files are not equal')

    def assertRegex(self, *args, **kwargs):
        ''' Compatibility across unittest versions '''
        if hasattr(unittest.TestCase, 'assertRegex'):
            return unittest.TestCase.assertRegex(self, *args, **kwargs)
        return self.assertRegexpMatches(*args, **kwargs)

    def assertNotEqual(self, *args, **kwargs):
        ''' Compatibility across unittest versions '''
        if hasattr(unittest.TestCase, 'assertNotEqual'):
            return unittest.TestCase.assertNotEqual(self, *args, **kwargs)
        return self.assertNotEquals(*args, **kwargs)

    def assertEqual(self, *args, **kwargs):
        ''' Compatibility across unittest versions '''
        if hasattr(unittest.TestCase, 'assertEqual'):
            return unittest.TestCase.assertEqual(self, *args, **kwargs)
        return self.assertEquals(*args, **kwargs)


def get_data_dir():
    ''' Return the testing data directory '''
    return os.path.join(os.path.dirname(__file__), 'data')


def file_contents(*args, encoding='utf-8'):
    ''' Return the contents of the file at the given path '''
    with open(os.path.join(*args), 'rb') as infile:
        return infile.read().decode('utf-8') 


def get_user_pass():
    ''' 
    Get the username and password from the environment if possible 

    If the environment does not contain a username and password,
    they will be retrieved from a ~/.authinfo file.

    '''
    username = None
    password = None
    if 'ESPUSER' in os.environ:
        username = os.environ['ESPUSER'] 
    if 'ESPPASSWORD' in os.environ:
        password = os.environ['ESPPASSWORD'] 
    return username, password


def get_host_port_proto():
    ''' 
    Get the host, port and protocol from a .esprc file 

    NOTE: .esprc files are written in Lua

    Returns
    -------
    (esphost, espport, espprotocol)

    '''
    esphost = os.environ.get('ESPHOST')
    espport = os.environ.get('ESPPORT')
    espprotocol = os.environ.get('ESPPROTOCOL')

    if espport is not None:
        espport = int(espport)

    if esphost and espport:
        return esphost, espport, espprotocol

    # If there is no host or port in the environment, look for .esprc
    esprc = None
    rcname = '.esprc'
    homepath = os.path.join(os.path.expanduser(os.environ.get('HOME', '~')), rcname)
    upath = os.path.join(r'u:', rcname)
    cfgfile = os.path.abspath(os.path.normpath(rcname))

    while not os.path.isfile(cfgfile):
        if os.path.samefile(os.path.dirname(homepath), os.path.dirname(cfgfile)):
            break
        newcfgfile = os.path.abspath(os.path.normpath(rcname)) 
        if os.path.samefile(os.path.dirname(cfgfile), os.path.dirname(newcfgfile)):
            break

    if os.path.isfile(cfgfile):
        esprc = cfgfile
    elif os.path.exists(homepath):
        esprc = homepath
    elif os.path.exists(upath):
        esprc = upath
    else:
        return esphost, espport, espprotocol

    return _read_esprc(esprc)


def _read_esprc(path):
    '''
    Read the .esprc file using Lua 

    Parameters
    ----------
    path : string
        Path to the .esprc file

    Returns
    -------
    (esphost, espport, espprotocol)

    '''
    esphost = None
    espport = None
    espprotocol = None

    if not os.path.isfile(path):
        return esphost, espport, espprotocol

    try:
        from lupa import LuaRuntime
        lua = LuaRuntime()
        lua.eval('dofile("%s")' % path)
        lg = lua.globals()

    except ImportError:
        import subprocess
        import tempfile

        lua_script = tempfile.TemporaryFile(mode='w')
        lua_script.write('''
            if arg[1] then
                dofile(arg[1])
                for name, value in pairs(_G) do
                    if name:match('esp') then
                        print(name .. ' ' .. tostring(value))
                    end
                end
            end
        ''')
        lua_script.seek(0)

        class LuaGlobals(object):
            pass

        lg = LuaGlobals()

        config = None
        try:
            config = subprocess.check_output('lua - %s' % path, stdin=lua_script,
                                             shell=True).strip().decode('utf-8')
        except (OSError, IOError, subprocess.CalledProcessError):
            pass
        finally:
            lua_script.close()

        if config:
            for name, value in re.findall(r'^\s*(esp\w+)\s+(.+?)\s*$', config, re.M):
                setattr(lg, name, value)
        else:
            # Try to parse manually
            config = re.sub(r'\-\-.*?$', r' ', open(path, 'r').read(), flags=re.M)
            for name, value in re.findall(r'\b(esp\w+)\s*=\s*(\S+)(?:\s+|\s*$)', config):
                setattr(lg, name, eval(value))

    try:
       esphost = str(lg.esphost)
    except:
       sys.sterr.write('ERROR: Could not access esphost setting\n')
       sys.exit(1)

    try:
       espport = int(lg.espport)
    except:
       sys.sterr.write('ERROR: Could not access espport setting\n')
       sys.exit(1)

    try:
       if lg.espprotocol:
           espprotocol = str(lg.espprotocol)
    except:
       pass

    return esphost, espport, espprotocol


def runtests(xmlrunner=False):
   ''' Run unit tests '''
   import sys

   if '--profile' in sys.argv:
       import profile
       import pstats

       sys.argv = [x for x in sys.argv if x != '--profile']

       if xmlrunner:
           import xmlrunner as xr
           profile.run("unittest.main(testRunner=xr.XMLTestRunner(output='test-reports', verbosity=2))", '_stats.txt')
       else:
           profile.run('unittest.main()', '_stats.txt')

       stats = pstats.Stats('_stats.txt')
       #stats.strip_dirs()
       stats.sort_stats('cumulative', 'calls')
       stats.print_stats(25)
       stats.sort_stats('time', 'calls')
       stats.print_stats(25)

   elif xmlrunner:
       import xmlrunner as xr
       unittest.main(testRunner=xr.XMLTestRunner(output='test-reports', verbosity=2)) 

   else:
       unittest.main()


def normalize_xml(s):
    # Remove http-servers
    s = re.sub(r'<http-servers>.*?</http-servers>', r'', s, flags=re.S)

    # Normalize connector type
    s = re.sub(r'(<connector[^>]+)(>\s*<properties>)\s*<property\s+name=[\'"]type[\'"]>(pub(?:lish)?|sub(?:scribe)?)</property>', r'\1 type="\3"\2', s)

    # Use full name of type
    s = re.sub(r'(type=[\'"])pub([\'"])', r'\1publish\2', s)
    s = re.sub(r'(type=[\'"])sub([\'"])', r'\1subscribe\2', s)

    # Remove CDATA from properties
    s = re.sub(r'(<property[^>]+>)<!\[CDATA\[(.+?)\]\]>(</property>)', r'\1\2\3', s)

    # Remove port= / copy-keys=
    s = re.sub(r'\s*(port|copy-keys)=[\'"][^\'"]*[\'"]\s*', r'', s)

    # Remove default attributes
    s = re.sub(r'\s*connection-timeout=[\'"]300[\'"]', r'', s)
    s = re.sub(r'\s*max-string-length=[\'"]32000[\'"]', r'', s)

    # Remove key='false'
    s = re.sub(r'\s+key=[\'"]false[\'"]', r'', s)

    # Remove engine / edges description
    s = re.sub(r'(<(engine|edges)[^>]*>\s*)<description>.*?</description>',
               r'\1', s, flags=re.S)

    # Normalize description whitespace
    s = re.sub(r'(<description>)\s*(.*?)\s*(</description>)',
               r'\1\2\3', s, flags=re.S)

    # Remove full path of files
    s = re.sub(r'(<property name=[\'"]fsname[\'"]>).*?(\w+\.\w+)(</property>)', r'\1\2\3', s)
    s = re.sub(r'(<code-file>).*?(\w+\.\w+)(</code-file>)', r'\1\2\3', s)

    # Expand schema strings
    s = re.sub(r'<schema-string>(.*?)</schema-string>', schema_string_to_schema, s)

    # Expand edges
    s = re.sub(r'(<edge [^>]*/>)', expand_edge, s)

    # Add engine / projects wrapper
    if re.match(r'^\s*<project\b', s):
        s = '<engine>\n<projects>\n%s\n</projects>\n</engine>\n' % s

    return s


def strip(s):
    return (s or '').strip()


def expand_edge(edge):
    out = []
    edge = xml.ensure_element(edge.group(1))
    sources = re.split(r'\s+', edge.attrib['source'].strip())
    targets = re.split(r'\s+', edge.attrib['target'].strip())
    for src in sources:
        for tgt in targets:
            new_elem = copy.deepcopy(edge)
            new_elem.attrib['source'] = src
            new_elem.attrib['target'] = tgt
            out.append(xml.to_xml(new_elem))
    return '\n'.join(out) 


def schema_string_to_schema(schema):
    schema = schema.group(1)
    out = ['<schema>', '<fields>']
    for item in re.split(r'\s*,\s*', schema.strip()):
        name, dtype = item.split(':')
        is_key = '*' in name and 'true' or 'false'
        name = name.replace('*', '')
        out.append('<field name="%s" type="%s" key="%s" />' % (name, dtype, is_key))
    out.append('</fields>')
    out.append('</schema>')
    return '\n'.join(out)


def elements_equal(e1, e2, path='/'):
    '''
    Compare elements (in an order-independent manner)

    Parameters
    ----------
    e1 : Element
        The first element
    e2 : Element
        The second element

    Raises
    ------
    ValueError
        Any difference raises a ValueError

    '''
    if isinstance(e1, six.string_types):
        e1 = xml.ensure_element(normalize_xml(e1))
    if isinstance(e2, six.string_types):
        e2 = xml.ensure_element(normalize_xml(e2))

    if e1.tag != e2.tag:
        raise ValueError('%s%s != %s%s' % (path, e1.tag, path, e2.tag))

    if strip(e1.text) != strip(e2.text):
        raise ValueError('In %s%s: %s != %s' % (path, e1.tag, strip(e1.text), strip(e2.text)))

    if strip(e1.tail) != strip(e2.tail):
        raise ValueError('In %s%s: %s != %s' % (path, e1.tag, strip(e1.tail), strip(e2.tail)))

    if e1.attrib != e2.attrib:
        raise ValueError('In %s%s: %s != %s' % (path, e1.tag, e1.attrib, e2.attrib))

    e1_children = sorted(list(e1), key=lambda x: (x.tag, x.attrib.get('name', ''), x.attrib.get('source', ''), x.attrib.get('target', ''), x.attrib.get('connector')))
    e2_children = sorted(list(e2), key=lambda x: (x.tag, x.attrib.get('name', ''), x.attrib.get('source', ''), x.attrib.get('target', ''), x.attrib.get('connector', '')))

    e1_tags = [x.tag for x in e1_children]
    e2_tags = [x.tag for x in e2_children]

    if e1_tags != e2_tags:
        print(xml.to_xml(e1, pretty=True))
        print(xml.to_xml(e2, pretty=True))
        raise ValueError('In %s%s: %s != %s' % (path, e1.tag, e1_tags, e2_tags))

    for c1, c2 in zip(e1_children, e2_children):
        elements_equal(c1, c2, path='%s%s/' % (path, e1.tag))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Schema '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import json
import os
import re
import requests
import six
import xml.etree.ElementTree as ET
from collections import OrderedDict
from six.moves import urllib
from .base import ESPObject, attribute
from .utils import xml


def clean_dtype(value):
    '''
    Make sure dtype is normalized

    '''
    return re.sub(r'\s+', r'', value)\
             .replace('array(double)', 'array(dbl)')\
             .replace('array(int32)', 'array(i32)')\
             .replace('array(int64)', 'array(i64)')


class FieldDict(collections.MutableMapping):
    ''' Dictionary for holding schema fields '''

    def __init__(self):
        self._data = OrderedDict()

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, name, value):
        key = False
        if '*' in name:
            name = name.replace('*', '')
            key = True

        # obj[name] = SchemaField(...)
        if isinstance(value, SchemaField):
            pass

        # obj[name] = 'type', True|False
        elif isinstance(value, (tuple, list)):
            if len(value) == 1:
                value = SchemaField(name, type=value[0], key=key)
            elif len(value) == 2:
                dtype, key = value
                if isinstance(key, six.string_types):
                    key = (key == 'true') and True or False
                value = SchemaField(name, type=clean_dtype(dtype), key=bool(key))
            else:
                raise ValueError('Too many values in schema field: %s' % value)

        # obj[name] = 'type'
        # obj['*name'] = 'type'
        # obj['name*'] = 'type'
        # obj[name] = '*type'
        # obj[name] = 'type*'
        elif isinstance(value, six.string_types):
            if '*' in value:
                value = SchemaField(name, type=value.replace('*'), key=True)
            else:
                value = SchemaField(name, type=value, key=key)

        # obj[name] = dict(type='type', key=True|False)
        elif isinstance(value, dict):
            value = SchemaField(name, value['type'], value.get('key', key))

        # obj[name] = ?
        else:
            raise TypeError('Unrecognized type for schema field: %s' % value)

        self._data[name] = value

    @property
    def schema_string(self):
        '''
        Construct schema string

        Returns
        -------
        string

        '''
        out = []
        for name, value in self.items():
            out.append('%s%s:%s' % (name, value.key and '*' or '', value.type))
        return ','.join(out)

    @schema_string.setter
    def schema_string(self, value):
        self.clear()
        for field in [x.strip() for x in re.split(r'\s*,\s*|\s+', value.strip())]:
            parts = list(field.split(':', 1))
            if len(parts) == 1:
                name = parts[0]
                dtype = 'inherit'
            else:
                name = parts[0]
                dtype = clean_dtype(parts[1])
            key = False
            if '*' in name:
                key = True
                name = name.replace('*', '')
            self[name] = SchemaField(name.strip(), dtype.strip(), key=key)

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)


class SchemaField(ESPObject):
    '''
    Schema field

    Attributes
    ----------
    name : string
        Name of the field
    type : string
        Data type of the field
    key : bool
        Indicates whether or not the field is a key field

    Parameters
    ----------
    name : string
        Name of the field
    type : string
        Data type of the field
    key : bool
        Indicates whether or not the field is a key field

    '''
    type = attribute('type', dtype='string')
    key = attribute('key', dtype='bool')

    def __init__(self, name, type, key=False):
        ESPObject.__init__(self, attrs=dict(type=type, key=key))
        self.name = name

    def copy(self, deep=False):
        '''
        Return a copy of the object

        Parameters
        ----------
        deep : bool, optional
            Should sub-objects be copied as well?

        Returns
        -------
        :class:`SchemaField`

        '''
        return type(self)(name=self.name, type=self.type, key=self.key)

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memo=None):
        return self.copy(deep=True)

    @classmethod
    def from_xml(cls, data, session=None):
        '''
        Create schema field from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            The schema field XML definition
        session : requests.Session
            The ESP session object

        Returns
        -------
        :class:`SchemaField`

        '''
        out = cls('', type='double', key=False)
        out.session = session

        if isinstance(data, six.string_types):
            data = ET.fromstring(data)

        out._set_attributes(data.attrib)

        return out

    from_element = from_xml

    def to_element(self):
        '''
        Export schema field definition to ElementTree.Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('field', xml.get_attrs(self))

    def to_xml(self, pretty=False):
        '''
        Export schema field definition to XML

        Parameters
        ----------
        pretty : bool, optional
            Should the output include whitespace for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def __str__(self):
        return '%s%s:%s' % (self.name, self.key and '*' or '', self.type)

    def __repr__(self):
        return str(self)


class Schema(ESPObject, collections.MutableMapping):
    '''
    Schema definition

    Attributes
    ----------
    fields : FieldDict
        Dictionary of field in the schema

    Parameters
    ----------
    *args : SchemaFields
        One or more :class:`SchemaField` objects
    name : string, optional
        Name for the schema object
    copy_window : string, optional
        Name of window to copy the scheme of
    copy_keys : boolean, optional
        If true, uses the keys from the schema that is copied with
        the copy=window_name attribute.
        Default: false

    '''

    name = attribute('name', dtype='string')
    copy_window = attribute('copy', dtype='string')
    copy_keys = attribute('copy-keys', dtype='bool')

    def __init__(self, *args, **kwargs):
        ESPObject.__init__(self)
        self.name = kwargs.get('name')
        self.copy_window = kwargs.get('copy_window')
        self.copy_keys = kwargs.get('copy_keys')
        self.fields = FieldDict()
        for item in args:
            self.fields[item.name] = item

    def copy(self, deep=False):
        '''
        Return a copy of the object

        Parameters
        ----------
        deep : bool, optional
            Should sub-objects be copied as well?

        Returns
        -------
        :class:`Schema`

        '''
        out = type(self)()
        for name, field in self.fields.items():
            if deep:
                out.fields[name] = field.copy(deep=True)
            else:
                out.fields[name] = field
        return out

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memo):
        return self.copy(deep=True)

    @property
    def schema_string(self):
        '''
        Construct schema string

        Returns
        -------
        string

        '''
        return self.fields.schema_string

    @schema_string.setter
    def schema_string(self, value):
        self.fields.schema_string = value

    @classmethod
    def from_json(cls, data, session=None):
        '''
        Create schema from JSON definition

        Parameters
        ----------
        data : json-string
            The schema definition in JSON
        session : requests.Session
            The ESP session object

        Returns
        -------
        :class:`Schema`

        '''
        out = cls()
        out.session = session

        fields = json.loads(data)['schema'][0]['fields']
        for field in fields:
            field = field['field']['attributes']
            key = field.get('key', 'false') == 'true'
            name = field['name']
            dtype = clean_dtype(field['type'])
            out.fields[name] = SchemaField(name, dtype, key=key)

        return out

    @classmethod
    def from_string(cls, data, session=None):
        '''
        Create schema from string definition

        Parameters
        ----------
        data : string
            The string schema definition
        session : requests.Session
            The ESP session object

        Returns
        -------
        :class:`Schema`

        '''
        out = cls()
        out.session = session

        for item in re.split(r'\s*,\s*', data.strip()):
            name, dtype = item.split(':')
            is_key = '*' in name
            name = name.replace('*', '')
            field = SchemaField(name, dtype, key=is_key)
            out.fields[name] = field

        return out

    @classmethod
    def from_schema_string(cls, data, session=None):
        '''
        Create schema from schema-string element

        Parameters
        ----------
        data : string
            The string-schema element
        session : requests.Session
            The ESP session object

        Returns
        -------
        :class:`Schema`

        '''
        if isinstance(data, six.string_types):
            if not data.strip().startswith('<'):
                data = '<schema-string>%s</schema-string>' % data
            data = xml.from_xml(data)

        if data.text:
            out = cls.from_string(data.text)
        else:
            out = cls()

        out.session = session

        out._set_attributes(data.attrib)

        return out

    @classmethod
    def from_xml(cls, data, session=None):
        '''
        Create schema from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            The schema XML definition
        session : requests.Session
            The ESP session object

        Returns
        -------
        :class:`Schema`

        '''
        out = cls()
        out.session = session

        if isinstance(data, six.string_types):
            data = xml.from_xml(data)

        out._set_attributes(data.attrib)

        for item in data.findall('.//field'):
            field = SchemaField.from_xml(item)
            out.fields[field.name] = field

        return out

    from_element = from_xml

    def to_element(self):
        '''
        Export schema definition to ElementTree.Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('schema', xml.get_attrs(self))

        if self.fields:
            fields = xml.add_elem(out, 'fields')
            for item in self.fields.values():
                xml.add_elem(fields, item.to_element())

        return out

    def to_xml(self, pretty=False):
        '''
        Export schema definition to XML

        Parameters
        ----------
        pretty : bool, optional
            Should the output include whitespace for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def __str__(self):
        return self.schema_string

    def __repr__(self):
        return str(self)

    def add_field(self, name, type, key=False):
        '''
        Add a schema field

        Parameters
        ----------
        name : string
            Name of the field
        type : string
            Data type of the field
        key : bool, optional
            Indicates whether or not the field is a key field

        '''
        self.fields[name] = SchemaField(name, type, key=key)

#
# MutableMapping methods
#

    def __getitem__(self, key):
        return self.fields[key]

    def __setitem__(self, key, value):
        self.fields[key] = value

    def __delitem__(self, key):
        del self.fields[key]

    def __iter__(self):
        return iter(self.fields)

    def __len__(self):
        return len(self.fields)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

import copy
from .base import Window, attribute
from .features import (SchemaFeature, CXXPluginContextFeature, CXXPluginFeature,
                       DS2TableServerFeature, DSExternalFeature, MASMapFeature)
from .utils import get_args, ensure_element, connectors_to_end
from ..utils import xml


class ProceduralWindow(Window, SchemaFeature, CXXPluginContextFeature,
                       CXXPluginFeature, DSExternalFeature):
    '''
    Procedural window

    Parameters
    ----------
    name : string, optional
        The name of the window
    schema : Schema, optional
        The schema of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level
        value of pubsub is manual, true enables publishing and subscribing
        for the window and false disables it.
    description : string, optional
        Description of the window
    output_insert_only : bool, optional
        When true, prevents the window from passing non-insert events
        to other windows.
    collapse_updates : bool, optional
        When true, multiple update blocks are collapsed into a single update block
    pulse_interval : string, optional
        Output a canonical batch of updates at the specified interval.
    exp_max_string : int, optional
        Specifies the maximum size of strings that the expression engine
        uses for the window. Default value is 1024.
    index_type : string, optional
        Index type for the window
    pubsub_index_type : string, optional
        Publish/subscribe index type.  Valid values are the same as for the
        `index_type` parameter.
    produces_only_inserts : bool, optional
        Set to true when you know that the Procedural window always
        produces inserts.

    Attributes
    ----------
    cxx_plugin_context : CXXPluginContext
        Specifies a shared library and function name.
    cxx_plugin : CXXPlugin
        A shared library and function name pair that specifies Procedural
        window handlers.
    ds_external : DSExternal
        Contains a block of DATA step code that is used as an input
        handler for Procedural windows.

    Returns
    -------
    :class:`ProceduralWindow`

    '''

    window_type = 'procedural'

    produces_only_inserts = attribute('produces-only-inserts', dtype='bool')

    def __init__(self, name=None, schema=None, pubsub=None,
                 description=None, output_insert_only=None,
                 collapse_updates=None, pulse_interval=None,
                 exp_max_string=None, index_type=None,
                 pubsub_index_type=None, produces_only_inserts=None):
        Window.__init__(self, **get_args(locals()))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

import six
from six.moves import urllib
from ..exceptions import ESPError
from ..utils import xml
from ..utils.xml import ensure_element


def verify_window(window, session=None):
    '''
    Verify that the window exists on the server

    Parameters
    ----------
    window : Window or string
        The window object or URL to check for
    session : requests.Session, optional
        The session to use to check for the window.  This is only used
        if the first parameter is a string.

    Returns
    -------
    boolean

    '''
    from .base import BaseWindow
    try:
        if isinstance(window, BaseWindow):
            window._get(urllib.parse.urljoin(window.base_url,
                                             'windows/%s/%s/%s' %
                                             (window.project,
                                              window.contquery,
                                              window.name)))
            return True
        elif isinstance(window, six.string_types):
            window = window.replace('.', '/')
            out = session.get(urllib.parse.urljoin(session.base_url,
                                                   'windows/%s' % window))
            if out.status_code < 300:
                return True
    except ESPError as exc:
        print(exc)
        pass
    return False


def get_args(kwargs):
    '''
    Return keyword arguments while removing `self`

    Parameters
    ----------
    kwargs : dict
        The keyword arguments dictionary

    Returns
    -------
    dict

    '''
    kwargs = dict(kwargs)
    kwargs.pop('self', None)
    return kwargs


def listify(data):
    '''
    Ensure data is a list

    Parameters
    ----------
    data : any
        The data to put into a list

    Returns
    -------
    list

    '''
    if data is None:
        return data
    if isinstance(data, (set, list, tuple)):
        return list(data)
    return [data]


def connectors_to_end(elem):
    ''' Move connectors to the end of element '''
    connectors = elem.find('./connectors')
    if connectors is not None:
        elem.remove(connectors)
        elem.append(connectors)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
import esppy
from . import utils as tm


class TestModelReaderWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('modelreader_window.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto',
                                  n_threads='1', use_tagged_token=True)

        cq = esp.ContinuousQuery(name='cq_01')

        w_source = esp.SourceWindow(name='w_source',
                                    schema=('Species:string', 'SepalLength:double',
                                            'SepalWidth:double', 'PetalLength:double',
                                            'PetalWidth:double', 'id*:int64'),
                                    insert_only=True, index_type='empty')

        w_request = esp.SourceWindow(name='w_request',
                                     schema=('req_id*:int64','req_key:string','req_val:string'),
                                     insert_only=True, index_type='empty')

        w_reader = esp.ModelReaderWindow(name='w_reader', pubsub=True)

        w_score = esp.ScoreWindow(name='w_score',
                                  schema=('id*:int64', 'SepalLength:double',
                                          'SepalWidth:double', 'PetalLength:double',
                                          'PetalWidth:double', 'Species:string',
                                          'P_SpeciesVersicolor:double',
                                          'P_SpeciesVirginica:double',
                                          'P_SpeciesSetosa:double',
                                          'I_Species:string', '_WARN_:string'))
        w_score.add_offline_model(model_type='astore',
                                  output_map=dict(test='testabc'))
        w_score.add_connector('fs',
                              conn_name='sub', conn_type='subscribe',
                              type='sub',
                              fstype='csv',
                              fsname='result.out',
                              snapshot=True)

        proj.add_query(cq)
        cq.add_windows(w_source, w_request, w_score, w_reader)

        w_source.add_target(w_score,role='data')
        w_reader.add_target(w_score,role='model')
        w_request.add_target(w_reader,role='request')

        self._test_model_file('modelreader_window.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP File and Socket Connectors '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class FilePublisher(Connector):
    '''
    Publish events from a file

    Parameters
    ----------
    fsname : string
        The filename or path of the file
    fstype : string, optional
        The data file type.
        Valid values: 'binary', 'csv', 'xml', 'json', 'syslog', 'hdat', 'cef'
    name : string, optional
        Name of the connector object
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    addcsvflags : string, optional
        Specifies the event type to insert into input CSV events.
        Valid values: 'normal' or 'partialupdate'
    addcsvopcode : string, optional
        Prepends an opcode and comma to input CSV events.
    blocksize : int, optional
        Specifies the number of events to include in a published event block
    cefsyslogprefix : string, optional
        When fstype=cef, specifies that CEF events contain the syslog prefix
    configfilesection : string, optional
        Specifies the name of the section in the ESP connector config
        file for parameters.
    csvfielddelimiter : string, optional
        Specifies the character delimiter for field data in input CSV events
    dateformat : string, optional
        Specifies the format of datetime and timestamp fields
    growinginputfile : boolean, optional
        Enables reading from a growing input file by publishers
    header : int, optional
        Specifies the number of input lines to skip before starting
        publish operations.
    ignorecsvparseerrors : boolean, optional
        Specifies that when a field in an input CSV event cannot be parsed,
        the event is dropped, an error is logged, and publishing continues.
    maxevents : int, optional
        Specifies the maximum number of events to publish
    noautogenfield : boolean, optional
        Specifies that input events are missing the key field that is
        autogenerated by the source window.
    prebuffer : boolean, optional
        Controls whether event blocks are buffered to an event block vector
        before doing any injects.
    publishwithupsert : boolean, optional
        Build events with opcode=upsert instead of insert
    rate : int, optional
        Specifies the requested transmit rate in events per second
    repeatcount : int, optional
        Specifies the number of times to repeat the publish operation
    transactional : string, optional
        Sets the event block type to transactional.

    Returns
    -------
    :class:`FilePublisher`

    '''
    connector_key = dict(cls='fs', type='publish')

    property_defs = dict(
        fsname=prop('fsname', dtype='string', required=True),
        fstype=prop('fstype', dtype='string', required=True,
                    default='csv',
                    valid_values=['binary', 'csv', 'xml', 'json',
                                  'syslog', 'hdat', 'cef']),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        addcsvflags=prop('addcsvflags', dtype='string',
                         valid_values=['normal', '']),
        addcsvopcode=prop('addcsvopcode', dtype='string',
                          valid_values=['insert', '']),
        blocksize=prop('blocksize', dtype='int', valid_expr='value > 0'),
        cefsyslogprefix=prop('blocksize', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        csvfielddelimiter=prop('csvfielddelimiter', dtype='string'),
        dateformat=prop('dateformat', dtype='string'),
        growinginputfile=prop('growinginputfile', dtype='boolean'),
        header=prop('header', dtype=('boolean', 'string'),
                    valid_values=[True, False, 'full']),
        ignorecsvparseerrors=prop('ignorecsvparseerrors', dtype='boolean'),
        maxevents=prop('maxevents', dtype='int', valid_expr='value >= 0'),
        noautogenfield=prop('noautogenfield', dtype='boolean'),
        prebuffer=prop('prebuffer', dtype='boolean'),
        publish_with_upsert=prop('publishwithupsert', dtype='boolean'),
        rate=prop('rate', dtype='int'),
        repeatcount=prop('repeatcount', dtype='int', valid_expr='value >= 0'),
        transactional=prop('transactional', dtype='string'),
    )

    def __init__(self, fsname=None, fstype=None, name=None, is_active=None,
                 snapshot=None, addcsvflags=None,
                 addcsvopcode=None, blocksize=None, cefsyslogprefix=None,
                 configfilesection=None, csvfielddelimiter=None,
                 dateformat=None,
                 growinginputfile=None, header=None, ignorecsvparseerrors=None,
                 maxevents=None, noautogenfield=None, prebuffer=None,
                 publishwithupsert=None, rate=None, repeatcount=None,
                 transactional=None):
        params = dict(**locals())
        params.pop('self')
        params.pop('is_active')
        name = params.pop('name')
        Connector.__init__(self, 'fs', name=name, type='publish', is_active=is_active,
                           properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties, required='fsname',
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)


class FileSubscriber(Connector):
    '''
    Subscribe to events from a file

    Parameters
    ----------
    fsname : string
        The filename or path of the file
    fstype : string, optional
        The data file type.
        Valid values: 'binary', 'csv', 'xml', 'json', 'syslog', 'hdat', 'cef'
    name : string, optional
        Name of the connector object
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    collapse : boolean, optional
        Converts UPDATE_BLOCK events to UPDATE events in order to make
        subscriber output publishable.
    configfilesection : string, optional
        Specifies the name of the section in the ESP connector config
        file for parameters.
    dateformat : string, optional
        Specifies the format of datetime and timestamp fields
    hdatcashostport : string, optional
        Specifies the CAS server host and port
    hdatcaspassword : string, optional
        Specifies the CAS server password
    hdatcasusername : string, optional
        Specifies the CAS server user name
    hdatfilename : string, optional
        Specifies the name of the Objective Analysis Package Data (HDAT)
        file to be written to the Hadoop Distributed File System (HDFS).
    hdatlasrhostport : string, optional
        Specifies the SAS LASR Analytic Server host and port
    hdatlasrkey : string, optional
        Specifies the path to tklasrkey.sh
    hdatmaxdatanodes : int, optional
        Specifies the maximum number of data node connections
    hdatmaxstringlength : int, optional
        Specifies in bytes the fixed size of string fields in Objective
        Analysis Package Data (HDAT) files
    hdatnumthreads : int, optional
        Specifies the size of the thread pool used for multi-threaded
        writes to data node socket connections.
    hdfsblocksize : int, optional
        Specifies in Mbytes the block size used to write an Objective
        Analysis Package Data (HDAT) file.
    hdfsnumreplicas : int, optional
        Specifies the number of Hadoop Distributed File System (HDFS)
        replicas created with writing an Objective Analysis Package
        Data (HDAT) file.
    header : boolean or string, optional
        For a CSV subscriber, specifies to write a header row that
        shows comma-separated fields.
        Valid values: True, False, or 'full' (include opcode flags in header)
    maxfilesize : int, optional
        Specifies the maximum size in bytes of the subscriber output file
    periodicity : int, optional
        Specifies the interval in seconds at which the subscriber output
        file is closed and a new output file opened.
    rate : boolean, optional
        When latency mode is enabled, shows this specified rate in generated
        output files.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks received
        by a subscriber that were introduced by a window retention policy.
    unbufferedoutputstreams : boolean, optional
        Specifies to create an unbuffered stream when writing to a
        file or socket.

    Returns
    -------
    :class:`FileSubscriber`

    '''
    connector_key = dict(cls='fs', type='subscribe')

    property_defs = dict(
        fsname=prop('fsname', dtype='string', required=True),
        fstype=prop('fstype', dtype='string', required=True,
                    default='csv',
                    valid_values=['binary', 'csv', 'xml', 'json',
                                  'syslog', 'hdat', 'cef']),
        snapshot=prop('snapshot', dtype='bool', required=True, default=False),
        collapse=prop('collapse', dtype='bool'),
        configfilesection=prop('configfilesection', dtype='string'),
        dateformat=prop('dateformat', dtype='string'),
        hdatcashostport=prop('hdatcashostport', dtype='string',
                             valid_values=re.compile(r'\w[\w\-\.]*:\d+')),
        hdatcaspassword=prop('hdatcaspassword', dtype='string'),
        hdatcasusername=prop('hdatcasusername', dtype='string'),
        hdatfilename=prop('hdatfilename', dtype='string'),
        hdatlasrhostport=prop('hdatlasrhostport', dtype='string'),
        hdatlasrkey=prop('hdatlasrkey', dtype='string'),
        hdatmaxdatanodes=prop('hdatmaxdatanodes',
                              dtype='int', valid_expr='value > 0'),
        hdatmaxstringlength=prop('hdatmaxstringlength', dtype='int',
                                 valid_expr='value > 0'),
        hdatnumthreads=prop('hdatnumthreads', dtype='int',
                            valid_expr='value >= 0'),
        hdfsblocksize=prop('hdfsblocksize', dtype='int',
                           valid_expr='value >= 0'),
        hdfsnumreplicas=prop('hdfsnumreplicas', dtype='int',
                             valid_expr='value >= 0'),
        header=prop('header', dtype=('bool', 'string'),
                    valid_values=[True, False, 'full']),
        maxfilesize=prop('maxfilesize', dtype='int',
                         valid_expr='value >= 0'),
        periodicity=prop('periodicity', dtype='int',
                         valid_expr='value >= 0'),
        rate=prop('rate', dtype='bool'),
        rmretdel=prop('rmretdel', dtype='bool'),
        unbufferedoutputstreams=prop('unbufferedoutputstreams', dtype='bool'),
    )

    def __init__(self, fsname=None, fstype=None, name=None, is_active=None,
                 snapshot=None, collapse=None, configfilesection=None,
                 dateformat=None, hdatcashostport=None,
                 hdatcaspassword=None, hdatcasusername=None,
                 hdatfilename=None, hdatlasrhostport=None, hdatlasrkey=None,
                 hdatmaxdatanodes=None, hdatmaxstringlength=None,
                 hdatnumthreads=None, hdfsblocksize=None,
                 hdfsnumreplicas=None, header=None, maxfilesize=None,
                 periodicity=None, rate=None, rmretdel=None,
                 unbufferedoutputstreams=None):
        params = dict(**locals())
        params.pop('self')
        params.pop('is_active')
        name = params.pop('name')
        Connector.__init__(self, 'fs', name=name, type='subscribe', is_active=is_active,
                           properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties, required='fsname',
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)


class SocketPublisher(FilePublisher):
    '''
    Publish events from a socket

    Parameters
    ----------
    port : int
        The port number to publish on
    fstype : string, optional
        The data file type.
        Valid values: 'binary', 'csv', 'xml', 'json', 'syslog', 'hdat', 'cef'
    name : string, optional
        Name of the connector object
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    addcsvflags : string, optional
        Specifies the event type to insert into input CSV events.
        Valid values: 'normal' or 'partialupdate'
    addcsvopcode : string, optional
        Prepends an opcode and comma to input CSV events.
    blocksize : int, optional
        Specifies the number of events to include in a published event block
    cefsyslogprefix : string, optional
        When fstype=cef, specifies that CEF events contain the syslog prefix
    configfilesection : string, optional
        Specifies the name of the section in the ESP connector config
        file for parameters.
    csvfielddelimiter : string, optional
        Specifies the character delimiter for field data in input CSV events
    dateformat : string, optional
        Specifies the format of datetime and timestamp fields
    growinginputfile : boolean, optional
        Enables reading from a growing input file by publishers
    header : int, optional
        Specifies the number of input lines to skip before starting
        publish operations.
    ignorecsvparseerrors : boolean, optional
        Specifies that when a field in an input CSV event cannot be parsed,
        the event is dropped, an error is logged, and publishing continues.
    maxevents : int, optional
        Specifies the maximum number of events to publish
    noautogenfield : boolean, optional
        Specifies that input events are missing the key field that is
        autogenerated by the source window.
    prebuffer : boolean, optional
        Controls whether event blocks are buffered to an event block vector
        before doing any injects.
    publishwithupsert : boolean, optional
        Build events with opcode=upsert instead of insert
    rate : int, optional
        Specifies the requested transmit rate in events per second
    repeatcount : int, optional
        Specifies the number of times to repeat the publish operation
    transactional : string, optional
        Sets the event block type to transactional.

    Returns
    -------
    :class:`SocketPublisher`

    '''
    connector_key = dict(cls='fs', type='publish', fsname=re.compile(r':(@\w+@|\d+)$'))

    def __init__(self, port=None, fstype=None, name=None, is_active=None,
                 snapshot=None, addcsvflags=None,
                 addcsvopcode=None, blocksize=None, cefsyslogprefix=None,
                 configfilesection=None, csvfielddelimiter=None,
                 dateformat=None,
                 growinginputfile=None, header=None, ignorecsvparseerrors=None,
                 maxevents=None, noautogenfield=None, prebuffer=None,
                 publishwithupsert=None, rate=None, repeatcount=None,
                 transactional=None):
        params = dict(locals())
        params.pop('is_active')
        if isinstance(port, six.string_types):
            port = port.replace(':', '')
        if port is not None:
            params['fsname'] = ':%s' % int(port)
        params.pop('port', None)
        FilePublisher.__init__(**params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties, required='fsname',
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)


class SocketSubscriber(FileSubscriber):
    '''
    Subscribe to events from a socket

    Parameters
    ----------
    host : string
        The host name where the socket exists
    port : int
        The port number on the server
    fstype : string, optional
        The data file type.
        Valid values: 'binary', 'csv', 'xml', 'json', 'syslog', 'hdat', 'cef'
    name : string, optional
        Name of the connector object
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    collapse : boolean, optional
        Converts UPDATE_BLOCK events to UPDATE events in order to make
        subscriber output publishable.
    configfilesection : string, optional
        Specifies the name of the section in the ESP connector config
        file for parameters.
    dateformat : string, optional
        Specifies the format of datetime and timestamp fields
    hdatcashostport : string, optional
        Specifies the CAS server host and port
    hdatcaspassword : string, optional
        Specifies the CAS server password
    hdatcasusername : string, optional
        Specifies the CAS server user name
    hdatfilename : string, optional
        Specifies the name of the Objective Analysis Package Data (HDAT)
        file to be written to the Hadoop Distributed File System (HDFS).
    hdatlasrhostport : string, optional
        Specifies the SAS LASR Analytic Server host and port
    hdatlasrkey : string, optional
        Specifies the path to tklasrkey.sh
    hdatmaxdatanodes : int, optional
        Specifies the maximum number of data node connections
    hdatmaxstringlength : int, optional
        Specifies in bytes the fixed size of string fields in Objective
        Analysis Package Data (HDAT) files
    hdatnumthreads : int, optional
        Specifies the size of the thread pool used for multi-threaded
        writes to data node socket connections.
    hdfsblocksize : int, optional
        Specifies in Mbytes the block size used to write an Objective
        Analysis Package Data (HDAT) file.
    hdfsnumreplicas : int, optional
        Specifies the number of Hadoop Distributed File System (HDFS)
        replicas created with writing an Objective Analysis Package
        Data (HDAT) file.
    header : boolean or string, optional
        For a CSV subscriber, specifies to write a header row that
        shows comma-separated fields.
        Valid values: True, False, or 'full' (include opcode flags in header)
    maxfilesize : int, optional
        Specifies the maximum size in bytes of the subscriber output file
    periodicity : int, optional
        Specifies the interval in seconds at which the subscriber output
        file is closed and a new output file opened.
    rate : boolean, optional
        When latency mode is enabled, shows this specified rate in generated
        output files.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks received
        by a subscriber that were introduced by a window retention policy.
    unbufferedoutputstreams : boolean, optional
        Specifies to create an unbuffered stream when writing to a
        file or socket.

    Returns
    -------
    :class:`SocketSubscriber`

    '''
    connector_key = dict(cls='fs', type='subscribe',
                         fsname=re.compile(r'^(@\w+@|[\w+\.\-]+):(@\w+@|\d+)$'))

    def __init__(self, host=None, port=None, fstype=None, name=None, is_active=None,
                 snapshot=None, collapse=None, configfilesection=None,
                 dateformat=None, hdatcashostport=None,
                 hdatcaspassword=None, hdatcasusername=None,
                 hdatfilename=None, hdatlasrhostport=None, hdatlasrkey=None,
                 hdatmaxdatanodes=None, hdatmaxstringlength=None,
                 hdatnumthreads=None, hdfsblocksize=None, hdfsnumreplicas=None,
                 header=None, maxfilesize=None, periodicity=None,
                 rate=None, rmretdel=None, unbufferedoutputstreams=None):
        params = dict(locals())
        params.pop('is_active')
        if host is not None and port is not None:
            params['fsname'] = '%s:%s' % (host, int(port))
        params.pop('host', None)
        params.pop('port', None)
        FileSubscriber.__init__(**params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties, required='fsname',
                                         delete='type')
        if req[0] is not None:
            host, port = req[0].split(':', 1)
        return cls(host, port, name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP OPC-UA Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class OPCUASubscriber(Connector):
    '''
    Subscribe to OPC-UA operations

    Parameters
    ----------
    opcuaendpoint : string, optional
        Specifies the OPC-UA server endpoint (only the portion
        following opc.tcp://).
    opcuanamespaceuri : string, optional
        Specifies the OPC-UA server namespace URI. The default is the
        namespace at index=0.
    opcuausername : string, optional
        Specifies the OPC-UA user name. The default is none.
    opcuapassword : string, optional
        Specifies the OPC-UA password. The default is none.
    opcuanodeids : string, optional
        Specifies a comma-separated list of Node IDs to map to ESP window
        schema fields, in the form <identifier type>_<identifier>. The list
        size must be equal to the number of fields in the subscribed window
        schema. Window field names are in Node ID form by default.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file to
        parse for configuration parameters. Specify the value
        as [configfilesection].
    snapshot : boolean, optional
        Specifies whether to send snapshot data

    Returns
    -------
    :class:`OPCUASubscriber`

    '''
    connector_key = dict(cls='opcua', type='subscribe')

    property_defs = dict(
        opcuaendpoint=prop('opcuaendpoint', dtype='string'),
        opcuanamespaceuri=prop('opcuanamespaceuri', dtype='string'),
        opcuausername=prop('opcuausername', dtype='string'),
        opcuapassword=prop('opcuapassword', dtype='string'),
        opcuanodeids=prop('opcuanodeids', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
    )

    def __init__(self, opcuaendpoint=None, name=None, is_active=None,
                 opcuanamespaceuri=None, opcuausername=None,
                 opcuapassword=None, opcuanodeids=None,
                 configfilesection=None, snapshot=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'opcua', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        properties = map_properties(cls, properties, delete='type')
        return cls(name=name, is_active=is_active, **properties)


class OPCUAPublisher(Connector):
    '''
    Publish OPC-UA operations

    Parameters
    ----------
    opcuaendpoint : string, optional
        Specifies the OPC-UA server endpoint (only the portion
        following opc.tcp://).
    opcuanamespaceuri : string, optional
        Specifies the OPC-UA server namespace URI. The default is the
        namespace at index=0.
    opcuausername : string, optional
        Specifies the OPC-UA user name. The default is none.
    opcuapassword : string, optional
        Specifies the OPC-UA password. The default is none.
    opcuanodeids : string, optional
        Specifies a comma-separated list of Node IDs to map to ESP window
        schema fields, in the form <identifier type>_<identifier>. The list
        size must be equal to the number of fields in the subscribed window
        schema. Window field names are in Node ID form by default.
    publishinterval : int, optional
        Specifies an interval in seconds when current values of all nodes
        in the Source window schema are published. The default is to publish
        when one or more values changes.
    transactional : string, optional
        Sets the event block type to transactional. The default value is normal.
    blocksize : int, optional
        Specifies the number of events to include in a published event
        block. The default value is 1.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file to
        parse for configuration parameters. Specify the value
        as [configfilesection].
    publishwithupsert : boolean, optional
        Builds events with opcode=Upsert instead of Insert.
    maxevents : int, optional
        Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`OPCUAPublisher`

    '''
    connector_key = dict(cls='opcua', type='publish')

    property_defs = dict(
        opcuaendpoint=prop('opcuaendpoint', dtype='string'),
        opcuanamespaceuri=prop('opcuanamespaceuri', dtype='string'),
        opcuausername=prop('opcuausername', dtype='string'),
        opcuapassword=prop('opcuapassword', dtype='string'),
        opcuanodeids=prop('opcuanodeids', dtype='string'),
        publishinterval=prop('publishinterval', dtype='int'),
        transactional=prop('transactional', dtype='string'),
        blocksize=prop('blocksize', dtype='int'),
        configfilesection=prop('configfilesection', dtype='string'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, opcuaendpoint=None, name=None, is_active=None,
                 opcuanamespaceuri=None, opcuausername=None,
                 opcuapassword=None, opcuanodeids=None,
                 publishinterval=None, transactional=None,
                 blocksize=None, configfilesection=None,
                 publishwithupsert=None, maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'opcua', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        properties = map_properties(cls, properties, delete='type')
        return cls(name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' XML Utilities '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import re
import six
import warnings
import xml.etree.ElementTree as ET
from .keyword import keywordify
from ..base import ESPObject


def _cast_attrs(attrs):
    out = {}
    for name, value in (attrs or {}).items():
        if value is None:
            continue
        elif value is True:
            value = 'true'
        elif value is False:
            value = 'false'
        else:
            value = '%s' % value
        out[name.replace('_', '-')] = value
    return out


def new_elem(elem_name, attrib=None, text_content=None, **kwargs):
    '''
    Create a new element

    Parameters
    ----------
    elem_name : string
       The tag name of the element
    attrib : dict, optional
        The attributes to set
    text_content : string, optional
        The text content of the new element
    **kwargs : keyword arguments, optional
        Additional attributes as keyword arguments

    Returns
    -------
    :class:`ElementTree.Element`

    '''
    attrib = _cast_attrs(attrib)
    kwargs = _cast_attrs(kwargs)
    out = ET.Element(elem_name, attrib=attrib, **kwargs)
    if text_content is not None:
        out.text = '%s' % text_content
    return out


def add_elem(parent_elem, child_elem, attrib=None,
             text_content=None, **kwargs):
    '''
    Add a new element to the specified parent element

    Parameters
    ----------
    parent_elem : ElementTree.Element
        The parent element
    child_elem : string
        The name of an element or an XML fragment
    attrib : dict, optional
        The attributes to set
    text_content : string, optional
        The text content of the new element
    **kwargs : keyword arguments, optional
        Additional attributes as keyword arguments

    Returns
    -------
    :class:`ElementTree.Element`

    '''
    attrib = _cast_attrs(attrib)
    kwargs = _cast_attrs(kwargs)

    # child_elem is an Element
    if isinstance(child_elem, ET.Element):
        out = child_elem
        if attrib:
            for key, value in attrib.items():
                out.set(key, value)
        for key, value in kwargs.items():
            out.set(key, value)
        parent_elem.append(out)

    # child_elem is an XML fragment
    elif re.match(r'^\s*<', child_elem):
        out = ET.fromstring(child_elem)
        if attrib:
            for key, value in attrib.items():
                out.set(key, value)
        for key, value in kwargs.items():
            out.set(key, value)
        parent_elem.append(out)

    # child_elem is an element name
    else:
        out = ET.SubElement(parent_elem, child_elem, attrib=attrib, **kwargs)

    if text_content is not None:
        out.text = '%s' % text_content

    return out


def add_properties(elem, *args, **kwargs):
    '''
    Add a ``properties`` node to the given element

    Parameters
    ----------
    elem : ElementTree.Element
        The element to add properties to
    verbatim : boolean, optional
        Should property names be used verbatim?
    *args : two-element-tuples, optional
        Passed to dict constructor as properties
    **kwargs : keyword arguments, optional
        Passed to dict constructor as properties

    Returns
    -------
    :class:`ElementTree.Element`

    '''
    verbatim = kwargs.pop('verbatim', False)
    bool_as_int = kwargs.pop('bool_as_int', False)
    props = add_elem(elem, 'properties')
    for key, value in sorted(dict(*args, **kwargs).items()):
        if value is None:
            continue
        if isinstance(value, (list, tuple, set)):
            value = ','.join(value)
        elif value is True:
            if bool_as_int:
                value = '1'
            else:
                value = 'true'
        elif value is False:
            if bool_as_int:
                value = '0'
            else:
                value = 'false'
        add_elem(props, 'property',
                 dict(name=keywordify(key)), text_content=value)
    return props


def xml_indent(elem, level=0):
    '''
    Add whitespace to XML for pretty-printing

    Parameters
    ----------
    elem : ElementTree.Element
        The element to modify with whitespace
    level : int, optional
        The level of indent

    Returns
    -------
    ``None``

    '''
    i = '\n' + (level * '  ')
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + '  '
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            xml_indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def from_xml(data):
    '''
    Convert XML to ElementTree.Element

    Parameters
    ----------
    data : string
        The XML to parse

    Returns
    -------
    :class:`ElementTree.Element`

    '''
    try:
        return ET.fromstring(data)
    except:
        for i, line in enumerate(data.split('\n')):
            print(i+1, line)
        raise


def to_xml(elem, encoding=None, pretty=False):
    '''
    Export element to XML

    Parameters
    ----------
    elem : ElementTree.Element or xml-string
        The element to export
    encoding : string, optional
        The output encoding

    Returns
    -------
    string

    '''
    if isinstance(elem, six.string_types):
        elem = ET.fromstring(elem)

    # In-place editing!!
    if pretty:
        xml_indent(elem)

    if encoding is None:
        return ET.tostring(elem, encoding='utf-8').decode()

    return ET.tostring(elem, encoding=encoding)


def get_attrs(obj, extra=[], exclude=[]):
    '''
    Retrieve XML attributes from object

    If ``obj`` has an ``xml_map`` dictionary attribute, it indicates
    the object attr to xml attr mapping.

        class MyObject(object):
            xml_map = dict(object_attr='xml_attr',
                           same_name_object_attr='same_name_object_attr')

    Parameters
    ----------
    obj : object
        The object to get attributes from

    Returns
    -------
    dict

    '''
    if isinstance(exclude, six.string_types):
        exclude = [exclude]

    out = obj._get_attributes()

    if isinstance(extra, six.string_types):
        extra = [extra]

    if extra:
        for item in extra:
            out[item] = getattr(obj, item)

    if exclude:
        for item in exclude:
            out.pop(item, None)

    return {k: '%s' % v for k, v in out.items() if v is not None}


def ensure_element(data):
    '''
    Ensure the given object is an ElementTree.Element

    Parameters
    ----------
    data : string or Element

    Returns
    -------
    :class:`ElementTree.Element`

    '''
    if isinstance(data, six.string_types):
        return from_xml(data)
    return data



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Websocket Client '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import warnings
import six
try:
    import wsaccel
except ImportError:
    wsaccel = None
from six.moves import urllib
from ws4py.client.threadedclient import WebSocketClient as WS4PyWebSocketClient


try:
    if wsaccel is not None:
        wsaccel.patch_ws4py()
except Exception as exc:
    warnings.warn('Could not import wsaccel: %s' % exc, RuntimeWarning)


class WebSocketClient(WS4PyWebSocketClient):
    '''
    Websocket Client

    Parameters
    ----------
    url : string
        The URL to connect to
    on_open : function, optional
        Function to call when the websocket is connected
    on_close : function, optional
        Function to call when the websocket is closed
    on_message : function, optional
        Function to call when a message is received
    on_error : function, optional
        Function to call when an error occurs
    **kwargs : keyword-arguments, optional
        Extra parameters to the ws4py socket client

    Returns
    -------
    :class:`WebSocketClient`

    '''

    def __init__(self, url, on_open=None, on_close=None, on_message=None,
                 on_error=None, **kwargs):
        self.callbacks = dict(on_open=on_open, on_close=on_close,
                              on_message=on_message, on_error=on_error)
        WS4PyWebSocketClient.__init__(self, url, **kwargs)

    def received_message(self, message):
        '''
        Handle a message from the server

        Parameters
        ----------
        message : string
            The data from the server

        '''
        if self.callbacks.get('on_message'):
            return self.callbacks['on_message'](self, message.data.decode(message.encoding))

    def unhandled_error(self, error):
        '''
        Handle an error from the server

        Parameters
        ----------
        error : Exception
           The exception associated with the error

        '''
        if self.callbacks.get('on_error'):
            self.callbacks['on_error'](self, error)

    def opened(self):
        ''' Handle the opening of a web socket connection '''
        if self.callbacks.get('on_open'):
            self.callbacks['on_open'](self)

    def closed(self, code, reason=None):
        '''
        Handle the closing of a web socket connection

        Parameters
        ----------
        code : int
            The code the server was closed with
        reason : string, optional
            The reason the server connection was closed

        '''
        if self.callbacks.get('on_close'):
            self.callbacks['on_close'](self, code, reason=reason)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Project Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class ProjectPublisher(Connector):
    '''
    ESP window event publisher

    Parameters
    ----------
    srcproject : string
        Specifies the name of the source project
    srccontinuousquery : string
        Specifies the name of the source continuous query
    srcwindow : string
        Specifies the name of the Source window
    maxevents : int, optional
        Specifies the maximum number of events to publish
    configfilesection : string, optional
        Specifies the name of the section in the config file to parse
        for configuration parameters. Specify the value as [configfilesection].

    Returns
    -------
    :class:`ProjectPublisher`

    '''
    connector_key = dict(cls='project', type='publish')

    property_defs = dict(
        srcproject=prop('srcproject', dtype='string', required=True),
        srccontinuousquery=prop('srccontinuousquery', dtype='string', required=True),
        srcwindow=prop('srcwindow', dtype='string', required=True),
        maxevents=prop('maxevents', dtype='int'),
        configfilesection=prop('configfilesection', dtype='string')
    )

    def __init__(self, srcproject=None, srccontinuousquery=None, srcwindow=None,
                 maxevents=None, configfilesection=None, name=None, is_active=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'project', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['srcproject',
                                                   'srccontinuousquery',
                                                   'srcwindow'],
                                         delete='type')
        return cls(req[0], req[1], req[2], name=name, is_active=is_active, **properties)
